from typing import Any, Dict, Optional, Union
from typing_extensions import Literal

from fastapi import Depends, FastAPI, Request
from fastapi.testclient import TestClient
from pytest import fixture, mark
from supertokens_python import init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import session
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.asyncio import create_new_session
from supertokens_python.recipe.session.framework.fastapi import verify_session
from supertokens_python.recipe.session.utils import TokenTransferMethod
from tests.utils import (
    extract_info,
    get_st_init_args,
    setup_function,
    start_st,
    teardown_function,
)

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore

pytestmark = mark.asyncio


@fixture(scope="function")
async def app():
    fast = FastAPI()
    fast.add_middleware(get_middleware())

    @fast.post("/create")
    async def _create(request: Request):  # type: ignore
        body = await request.json()
        session = await create_new_session(request, "userId", body, {})
        return {"message": True, "sessionHandle": session.get_handle()}

    @fast.get("/update-payload")
    async def _update_payload(session: SessionContainer = Depends(verify_session())):  # type: ignore
        await session.merge_into_access_token_payload({"newKey": "test"})
        return {"message": True}

    @fast.get("/verify")
    async def _verify(session: SessionContainer = Depends(verify_session())):  # type: ignore
        return {
            "message": True,
            "sessionHandle": session.get_handle(),
            "sessionExists": True,
        }

    @fast.get("/verify-optional")
    async def _verify_optional(session: Optional[SessionContainer] = Depends(verify_session(session_required=False))):  # type: ignore
        return {
            "message": True,
            "sessionHandle": session.get_handle() if session is not None else None,
            "sessionExists": session is not None,
        }

    return TestClient(fast)


from requests.cookies import RequestsCookieJar


def call_api(
    app: TestClient,
    info: Dict[str, Any],
    url: str,
    expected_status: int,
    auth_mode: str,
    auth_mode_header: Optional[str] = None,
):
    access_token = info.get("accessToken") or info.get("accessTokenFromHeader")

    headers = {}
    cookies = {}

    if auth_mode_header:
        headers["st-auth-mode"] = auth_mode_header
    if auth_mode in ("cookie", "both"):
        cookies["sAccessToken"] = access_token
        if info.get("antiCsrf") is not None:
            headers["anti-csrf"] = info["antiCsrf"]

    if auth_mode in ("header", "both"):
        headers[
            "Authorization"
        ] = f"Bearer {(access_token)}"  # TODO: Might have to add decode_uri()

    app.cookies = RequestsCookieJar()  # Reset cookies

    res = app.get(
        url=url,
        headers=headers,
        cookies=cookies,
    )

    assert res.status_code == expected_status
    return res


EXAMPLE_JWT = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"


def create_session(
    app: TestClient,
    auth_mode_header: Optional[str] = None,
    body: Optional[Dict[str, Any]] = None,
    cookies: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
):
    if headers is None:
        headers = {}
    if cookies is None:
        cookies = {}

    if auth_mode_header is not None:
        headers["st-auth-mode"] = auth_mode_header

    res = app.post(url="/create", headers=headers, json=body or {}, cookies=cookies)

    return extract_info(res)


def check_extracted_info(
    res: Dict[str, Any],
    expected_transfer_method: TokenTransferMethod,
    passed_different_token: bool = False,
):
    if expected_transfer_method == "header":
        for prop in ["accessToken", "refreshToken"]:
            if (
                passed_different_token
            ):  # If method is header but we passed a different token in cookie with request
                assert (
                    res.get(prop, "") == ""
                )  # It should clear the cookie (if present)
            else:
                assert res[prop] is None
        assert res["antiCsrf"] is None
        for prop in ["accessTokenFromHeader", "refreshTokenFromHeader"]:
            assert res[prop] != ""

        # We check that we will have access token at least as long as we have a refresh token
        # so verify session can return TRY_REFRESH_TOKEN
        # assert res["accessTokenFromHeader"]["expiry"] >= res["refreshTokenFromHeader"]["expiry"]
    elif expected_transfer_method == "cookie":
        for prop in ["accessToken", "refreshToken", "antiCsrf"]:
            assert res[prop] != ""
        for prop in ["accessTokenFromHeader", "refreshTokenFromHeader"]:
            if (
                passed_different_token
            ):  # If method is cookie but we passed a different token in header with request
                assert res.get(prop, "") == ""  # clear the header (if present)
            else:
                assert res[prop] is None
        # We check that we will have access token at least as long as we have a refresh token
        # so verify session can return TRY_REFRESH_TOKEN
        # assert datetime.parse(res["accessTokenExpiry"]) >= datetime.parse(res["refreshTokenExpiry"])
    else:
        assert False, "Invalid expected_transfer_method"


@mark.parametrize(
    "auth_mode_header, expected_transfer_method",
    [
        (None, "header"),  # default
        ("invalid-token-transfer-method", "header"),  # invalid
        ("header", "header"),  # specified
        ("cookie", "cookie"),  #  specified
    ],
)
async def test_should_follow_auth_mode_header(
    app: TestClient,
    auth_mode_header: Optional[str],
    expected_transfer_method: TokenTransferMethod,
):
    init(**get_st_init_args([session.init(anti_csrf="VIA_TOKEN")]))  # type:ignore
    start_st()

    res = create_session(app, auth_mode_header)

    check_extracted_info(res, expected_transfer_method)


@mark.parametrize(
    "token_transfer_method, expected_transfer_method, cookies, headers",
    [
        ("any", "header", {}, {}),  # default
        ("header", "header", {"sAccessToken": EXAMPLE_JWT}, {}),  # specified
        ("cookie", "cookie", {}, {}),  #  specified
        (
            "cookie",
            "cookie",
            {},
            {"Authorization": f"Bearer {EXAMPLE_JWT}"},
        ),  #  specified
    ],
)
async def test_should_follow_get_token_transfer_method(
    app: TestClient,
    token_transfer_method: Union[TokenTransferMethod, Literal["any"]],
    expected_transfer_method: TokenTransferMethod,
    cookies: Dict[str, Any],
    headers: Dict[str, Any],
):
    init(
        **get_st_init_args(
            [
                session.init(
                    anti_csrf="VIA_TOKEN",
                    get_token_transfer_method=lambda _, __, ___: token_transfer_method,  # type: ignore
                )
            ]
        )
    )  # type:ignore
    start_st()

    res = create_session(app, None, None, cookies=cookies, headers=headers)

    passed_different_token = (len(cookies) != 0) or (len(headers) != 0)
    check_extracted_info(res, expected_transfer_method, passed_different_token)

    if len(cookies) != 0 and expected_transfer_method == "header":
        assert res["sAccessToken"]["value"] == ""
        assert res["sAccessToken"]["expires"] == "Thu, 01 Jan 1970 00:00:00 GMT"
        assert res["sRefreshToken"]["value"] == ""
        assert res["sRefreshToken"]["expires"] == "Thu, 01 Jan 1970 00:00:00 GMT"

    if len(headers) != 0 and expected_transfer_method == "cookie":
        assert res["accessTokenFromHeader"] == ""
        assert res["refreshTokenFromHeader"] == ""


@mark.parametrize(
    "transfer_method, session_required, auth_header, auth_cookie, result",
    [
        ("any", False, False, False, None),
        ("header", False, False, False, None),
        ("cookie", False, False, False, None),
        ("cookie", False, True, False, None),
        ("header", False, False, True, None),
        ("any", True, False, False, "UNAUTHORISED"),
        ("header", True, False, False, "UNAUTHORISED"),
        ("cookie", True, False, False, "UNAUTHORISED"),
        ("cookie", True, True, False, "UNAUTHORISED"),
        ("header", True, False, True, "UNAUTHORISED"),
        ("any", True, True, True, "validateheader"),
        ("any", False, True, True, "validateheader"),
        ("header", True, True, True, "validateheader"),
        ("header", False, True, True, "validateheader"),
        ("cookie", True, True, True, "validatecookie"),
        ("cookie", False, True, True, "validatecookie"),
        ("any", True, True, False, "validateheader"),
        ("any", False, True, False, "validateheader"),
        ("header", True, True, False, "validateheader"),
        ("header", False, True, False, "validateheader"),
        ("any", True, False, True, "validatecookie"),
        ("any", False, False, True, "validatecookie"),
        ("cookie", True, False, True, "validatecookie"),
        ("cookie", False, False, True, "validatecookie"),
    ],
)
def test_verify_session_parametrized(
    app: TestClient,
    transfer_method: TokenTransferMethod,
    session_required: bool,
    auth_header: bool,
    auth_cookie: bool,
    result: Optional[str],
):
    init(
        **get_st_init_args(
            [
                session.init(
                    anti_csrf="VIA_TOKEN",
                    get_token_transfer_method=lambda _, __, ___: transfer_method,
                )
            ]
        )
    )
    start_st()

    create_session_info = create_session(app, "cookie")

    # if use_expired_token:
    #     delay(3)

    auth_mode = "none"
    if auth_cookie and auth_header:
        auth_mode = "both"
    elif auth_header:
        auth_mode = "header"
    elif auth_cookie:
        auth_mode = "cookie"

    res = call_api(
        app,
        create_session_info,
        "/verify" if session_required else "/verify-optional",
        401 if result == "UNAUTHORISED" else 200,
        auth_mode,
        # FIXME: missing auth_mode_header?
    )
    assert res.status_code == (401 if result == "UNAUTHORISED" else 200)

    body = res.json()

    if result is None:
        assert body["sessionExists"] is False
    if result == "UNAUTHORISED":
        assert body["message"] == "unauthorised"
    if result in ("validateCookie", "validateHeader"):
        assert body["sessionExists"] is True


# # Skipping the following 3 tests as they are covered by previous tests:
# # Provide access token via both header and cookie. It should use the value from headers if getTokenTransferMethod returns header or any. If returns cookie, it should use the value from cookie.


async def test_should_reject_requests_with_sIdRefreshToken(app: TestClient):
    init(**get_st_init_args([session.init(anti_csrf="VIA_TOKEN")]))
    start_st()

    res = create_session(
        app,
        "cookie",
        None,
        cookies={"sIdRefreshToken": "IRRELEVANT-VALUE", "sAccessToken": EXAMPLE_JWT},
    )

    response = app.get(
        url="/verify",
        cookies={
            "sIdRefreshToken": "IRRELEVANT-VALUE",
            "sAccessToken": EXAMPLE_JWT,
        },
        headers={"anti-csrf": res["antiCsrf"]},
    )

    assert response.status_code == 401
    assert response.json() == {"message": "try refresh token"}


# # SKIPPING:
# # with non ST in Authorize header
# # should use the value from cookies if present and getTokenTransferMethod returns {any,header,cookie}


# # merge_into_access_token_payload


@mark.parametrize("transfer_method", ["header", "cookie"])
async def test_should_update_acccess_token_payload(
    app: TestClient, transfer_method: str
):
    init(**get_st_init_args([session.init(anti_csrf="VIA_TOKEN")]))
    start_st()

    res = create_session(app, transfer_method)

    update_info = extract_info(
        call_api(app, res, "/update-payload", 200, "cookie", None)
    )

    # Didn't update
    assert update_info["refreshToken"] is None
    assert update_info["antiCsrf"] is None
    assert update_info["accessTokenFromHeader"] is None
    assert update_info["refreshTokenFromHeader"] is None

    # Updated access token
    assert update_info["accessToken"] is not None
    assert update_info["accessToken"] != res["accessTokenFromHeader"]
    # Updated front token
    assert update_info["frontToken"] is not None
    assert update_info["frontToken"] != res["frontToken"]


# refresh_session


@mark.parametrize(
    "transfer_method, auth_header, auth_cookie, output, set_tokens, cleared_tokens",
    [
        ("any", False, False, "unauthorised", None, None),
        ("header", False, False, "unauthorised", None, None),
        ("cookie", False, False, "unauthorised", None, None),
        ("any", False, True, "validatecookie", "cookies", None),
        ("header", False, True, "unauthorised", None, None),
        ("cookie", False, True, "validatecookie", "cookies", None),
        ("any", True, False, "validateheader", "headers", None),
        ("header", True, False, "validateheader", "headers", None),
        ("cookie", True, False, "unauthorised", None, None),
        ("any", True, True, "validateheader", "headers", "cookies"),
        ("header", True, True, "validateheader", "headers", "cookies"),
        ("cookie", True, True, "validatecookie", "cookies", "headers"),
    ],
)
async def test_xx(  # test_refresh_session_parametrized
    app: TestClient,
    transfer_method: TokenTransferMethod,
    auth_header: bool,
    auth_cookie: bool,
    output: Optional[str],
    set_tokens: Optional[str],
    cleared_tokens: Optional[str],
):
    init(
        **get_st_init_args(
            [
                session.init(
                    anti_csrf="VIA_TOKEN",
                    get_token_transfer_method=lambda _, __, ___: transfer_method,
                )
            ]
        )
    )
    start_st()

    # Token transfer method doesn't matter for this test
    res = create_session(app, "cookies")

    auth_mode = "none"
    if auth_cookie and auth_header:
        auth_mode = "both"
    elif auth_header:
        auth_mode = "header"
    elif auth_cookie:
        auth_mode = "cookie"

    refresh_result = extract_info(
        await refresh_session(app, transfer_method, auth_mode, res)
    )

    if output == "unauthorised":
        assert refresh_result["status_code"] == 401
        assert refresh_result["body"] == {"message": "unauthorised"}
    else:
        assert refresh_result["status_code"] == 200

    if cleared_tokens == "headers":
        assert refresh_result["accessTokenFromHeader"] == ""
        assert refresh_result["refreshTokenFromHeader"] == ""
    elif cleared_tokens == "cookies":
        assert refresh_result["accessToken"] == ""
        assert (
            refresh_result["sAccessToken"]["expires"] == "Thu, 01 Jan 1970 00:00:00 GMT"
        )
        assert refresh_result["refreshToken"] == ""
        assert (
            refresh_result["sRefreshToken"]["expires"]
            == "Thu, 01 Jan 1970 00:00:00 GMT"
        )

    if set_tokens == "headers":
        assert refresh_result["accessTokenFromHeader"] != ""
        assert refresh_result["refreshTokenFromHeader"] != ""
    elif set_tokens == "cookies":
        assert refresh_result["sAccessToken"]["value"] != ""
        assert (
            refresh_result["sAccessToken"]["expires"] != "Thu, 01 Jan 1970 00:00:00 GMT"
        )
        assert refresh_result["sRefreshToken"]["value"] != ""
        assert (
            refresh_result["sRefreshToken"]["expires"]
            != "Thu, 01 Jan 1970 00:00:00 GMT"
        )
    elif set_tokens is None:
        if cleared_tokens is None:
            assert refresh_result["frontToken"] is None
        return
    else:
        assert False, "Invalid set_tokens value"

    if set_tokens != "cookies" and cleared_tokens != "cookies":
        assert refresh_result["accessToken"] is None
        assert refresh_result["refreshToken"] is None
    elif set_tokens != "headers" and cleared_tokens != "headers":
        assert refresh_result["accessTokenFromHeader"] is None
        assert refresh_result["refreshTokenFromHeader"] is None


async def refresh_session(
    app: TestClient,
    auth_mode_header: TokenTransferMethod,
    auth_mode: str,
    res: Dict[str, str],
):
    headers = {}
    cookies = {}

    app.cookies = RequestsCookieJar()  # Reset cookies

    if auth_mode_header:
        headers["st-auth-mode"] = auth_mode_header

    access_token = res["accessToken"] or res["accessTokenFromHeader"]
    refresh_token = res["refreshToken"] or res["refreshTokenFromHeader"]

    if auth_mode in ("both", "cookie"):
        cookies["sAccessToken"] = access_token
        cookies["sRefreshToken"] = refresh_token
        if res.get("antiCsrf"):
            headers["anti-csrf"] = res["antiCsrf"]

    if auth_mode in ("both", "header"):
        headers["authorization"] = f"Bearer {refresh_token}"

    return app.post("/auth/session/refresh", headers=headers, cookies=cookies)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Union, Callable, Optional, TypeVar

from supertokens_python.async_to_sync_wrapper import sync
from supertokens_python.recipe.openid.interfaces import (
    GetOpenIdDiscoveryConfigurationResult,
)
from supertokens_python.types import MaybeAwaitable

from ...jwt.interfaces import (
    CreateJwtOkResult,
    CreateJwtResultUnsupportedAlgorithm,
    GetJWKSResult,
)
from ..interfaces import (
    RegenerateAccessTokenOkResult,
    SessionContainer,
    SessionInformationResult,
    SessionClaimValidator,
    SessionClaim,
    JSONObject,
    ClaimsValidationResult,
    SessionDoesNotExistError,
    GetClaimValueOkResult,
)


def create_new_session(
    request: Any,
    user_id: str,
    access_token_payload: Union[Dict[str, Any], None] = None,
    session_data: Union[Dict[str, Any], None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> SessionContainer:
    from supertokens_python.recipe.session.asyncio import (
        create_new_session as async_create_new_session,
    )

    return sync(
        async_create_new_session(
            request,
            user_id,
            access_token_payload,
            session_data,
            user_context=user_context,
        )
    )


def get_session(
    request: Any,
    anti_csrf_check: Union[bool, None] = None,
    session_required: bool = True,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], SessionContainer, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[SessionContainer, None]:
    from supertokens_python.recipe.session.asyncio import (
        get_session as async_get_session,
    )

    return sync(
        async_get_session(
            request,
            anti_csrf_check,
            session_required,
            override_global_claim_validators,
            user_context,
        )
    )


def refresh_session(
    request: Any, user_context: Union[None, Dict[str, Any]] = None
) -> SessionContainer:
    from supertokens_python.recipe.session.asyncio import (
        refresh_session as async_refresh_session,
    )

    return sync(async_refresh_session(request, user_context))


def revoke_session(
    session_handle: str, user_context: Union[None, Dict[str, Any]] = None
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        revoke_session as async_revoke_session,
    )

    return sync(async_revoke_session(session_handle, user_context))


def revoke_all_sessions_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[str]:
    from supertokens_python.recipe.session.asyncio import (
        revoke_all_sessions_for_user as async_revoke_all_sessions_for_user,
    )

    return sync(async_revoke_all_sessions_for_user(user_id, user_context))


def get_all_session_handles_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[str]:
    from supertokens_python.recipe.session.asyncio import (
        get_all_session_handles_for_user as async_get_all_session_handles_for_user,
    )

    return sync(async_get_all_session_handles_for_user(user_id, user_context))


def revoke_multiple_sessions(
    session_handles: List[str], user_context: Union[None, Dict[str, Any]] = None
) -> List[str]:
    from supertokens_python.recipe.session.asyncio import (
        revoke_multiple_sessions as async_revoke_multiple_sessions,
    )

    return sync(async_revoke_multiple_sessions(session_handles, user_context))


def get_session_information(
    session_handle: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[SessionInformationResult, None]:
    from supertokens_python.recipe.session.asyncio import (
        get_session_information as async_get_session_information,
    )

    return sync(async_get_session_information(session_handle, user_context))


def update_session_data(
    session_handle: str,
    new_session_data: Dict[str, Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        update_session_data as async_update_session_data,
    )

    return sync(
        async_update_session_data(session_handle, new_session_data, user_context)
    )


def update_access_token_payload(
    session_handle: str,
    new_access_token_payload: Dict[str, Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        update_access_token_payload as async_update_access_token_payload,
    )

    return sync(
        async_update_access_token_payload(
            session_handle, new_access_token_payload, user_context
        )
    )


def merge_into_access_token_payload(
    session_handle: str,
    new_access_token_payload: Dict[str, Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        merge_into_access_token_payload as async_merge_into_access_token_payload,
    )

    return sync(
        async_merge_into_access_token_payload(
            session_handle, new_access_token_payload, user_context
        )
    )


def create_jwt(
    payload: Dict[str, Any],
    validity_seconds: Union[None, int] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[CreateJwtOkResult, CreateJwtResultUnsupportedAlgorithm]:
    from supertokens_python.recipe.session.asyncio import create_jwt as async_create_jwt

    return sync(async_create_jwt(payload, validity_seconds, user_context))


def get_jwks(user_context: Union[None, Dict[str, Any]] = None) -> GetJWKSResult:
    from supertokens_python.recipe.session.asyncio import get_jwks as async_get_jwks

    return sync(async_get_jwks(user_context))


def get_open_id_discovery_configuration(
    user_context: Union[None, Dict[str, Any]] = None
) -> GetOpenIdDiscoveryConfigurationResult:
    from supertokens_python.recipe.session.asyncio import (
        get_open_id_discovery_configuration as async_get_open_id_discovery_configuration,
    )

    return sync(async_get_open_id_discovery_configuration(user_context))


def regenerate_access_token(
    access_token: str,
    new_access_token_payload: Union[Dict[str, Any], None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[RegenerateAccessTokenOkResult, None]:
    from supertokens_python.recipe.session.asyncio import (
        regenerate_access_token as async_regenerate_access_token,
    )

    return sync(
        async_regenerate_access_token(
            access_token, new_access_token_payload, user_context
        )
    )


_T = TypeVar("_T")


def fetch_and_set_claim(
    session_handle: str,
    claim: SessionClaim[Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        fetch_and_set_claim as async_fetch_and_set_claim,
    )

    return sync(async_fetch_and_set_claim(session_handle, claim, user_context))


def set_claim_value(
    session_handle: str,
    claim: SessionClaim[_T],
    value: _T,
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        set_claim_value as async_set_claim_value,
    )

    return sync(async_set_claim_value(session_handle, claim, value, user_context))


def get_claim_value(
    session_handle: str,
    claim: SessionClaim[_T],
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[SessionDoesNotExistError, GetClaimValueOkResult[_T]]:
    from supertokens_python.recipe.session.asyncio import (
        get_claim_value as async_get_claim_value,
    )

    return sync(async_get_claim_value(session_handle, claim, user_context))


def remove_claim(
    session_handle: str,
    claim: SessionClaim[Any],
    user_context: Union[None, Dict[str, Any]] = None,
) -> bool:
    from supertokens_python.recipe.session.asyncio import (
        remove_claim as async_remove_claim,
    )

    return sync(async_remove_claim(session_handle, claim, user_context))


def validate_claims_for_session_handle(
    session_handle: str,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], SessionInformationResult, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[SessionDoesNotExistError, ClaimsValidationResult]:
    from supertokens_python.recipe.session.asyncio import (
        validate_claims_for_session_handle as async_validate_claims_for_session_handle,
    )

    return sync(
        async_validate_claims_for_session_handle(
            session_handle, override_global_claim_validators, user_context
        )
    )


def validate_claims_in_jwt_payload(
    user_id: str,
    jwt_payload: JSONObject,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], str, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from supertokens_python.recipe.session.asyncio import (
        validate_claims_in_jwt_payload as async_validate_claims_in_jwt_payload,
    )

    return sync(
        async_validate_claims_in_jwt_payload(
            user_id, jwt_payload, override_global_claim_validators, user_context
        )
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Union

from supertokens_python.recipe.thirdparty.recipe import ThirdPartyRecipe

from ..types import User


async def get_user_by_id(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[User, None]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyRecipe.get_instance().recipe_implementation.get_user_by_id(
        user_id, user_context
    )


async def get_users_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[User]:
    if user_context is None:
        user_context = {}
    return (
        await ThirdPartyRecipe.get_instance().recipe_implementation.get_users_by_email(
            email, user_context
        )
    )


async def get_user_by_third_party_info(
    third_party_id: str,
    third_party_user_id: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyRecipe.get_instance().recipe_implementation.get_user_by_thirdparty_info(
        third_party_id, third_party_user_id, user_context
    )


async def sign_in_up(
    third_party_id: str,
    third_party_user_id: str,
    email: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyRecipe.get_instance().recipe_implementation.sign_in_up(
        third_party_id, third_party_user_id, email, user_context
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from supertokens_python.recipe.thirdparty.utils import find_right_provider

if TYPE_CHECKING:
    from supertokens_python.recipe.thirdparty.interfaces import APIOptions, APIInterface

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.utils import default_user_context, send_200_response


async def handle_sign_in_up_api(
    api_implementation: APIInterface, api_options: APIOptions
):
    if api_implementation.disable_sign_in_up_post:
        return None
    body = await api_options.request.json()
    if body is None:
        raise_bad_input_exception("Please provide a JSON input")

    code = body["code"] if "code" in body else ""
    auth_code_response = (
        body["authCodeResponse"] if "authCodeResponse" in body else None
    )
    client_id = body["clientId"] if "clientId" in body else None

    if "thirdPartyId" not in body or not isinstance(body["thirdPartyId"], str):
        raise_bad_input_exception("Please provide the thirdPartyId in request body")

    if not isinstance(code, str):
        raise_bad_input_exception(
            "Please make sure that the code in the request body is a string"
        )

    if code == "" and auth_code_response is None:
        raise_bad_input_exception(
            "Please provide one of code or authCodeResponse in the request body"
        )

    if auth_code_response is not None and "access_token" not in auth_code_response:
        raise_bad_input_exception(
            "Please provide the access_token inside the authCodeResponse request param"
        )

    if "redirectURI" not in body or not isinstance(body["redirectURI"], str):
        raise_bad_input_exception("Please provide the redirectURI in request body")

    third_party_id = body["thirdPartyId"]
    provider = find_right_provider(api_options.providers, third_party_id, client_id)
    if provider is None:
        if client_id is None:
            raise_bad_input_exception(
                "The third party provider "
                + third_party_id
                + " seems to be missing from the backend configs."
            )
        raise_bad_input_exception(
            "The third party provider "
            + third_party_id
            + " seems to be missing from the backend configs. If it is configured, then please make sure that you are passing the correct clientId from the frontend."
        )
    user_context = default_user_context(api_options.request)

    result = await api_implementation.sign_in_up_post(
        provider,
        code,
        body["redirectURI"],
        client_id,
        auth_code_response,
        api_options,
        user_context,
    )
    return send_200_response(result.to_json(), api_options.response)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from typing import Any, Dict

from fastapi import Response


def extract_all_cookies(response: Response) -> Dict[str, Any]:
    cookie_headers = response.headers.getlist("Set-Cookie")
    cookies: Dict[str, Any] = {}
    for header in cookie_headers:
        attributes = header.split(";")
        cookie = {}
        is_name = True
        name: str = ""
        for attr in attributes:
            split = attr.split("=")
            if is_name:
                name = split[0].strip()
                cookie["value"] = split[1]
                is_name = False
            else:
                cookie[split[0].strip().lower()] = split[1] if len(split) > 1 else True
        cookies[name] = cookie
    return cookies



from pytest import fixture, mark
from fastapi import FastAPI
from supertokens_python.framework.fastapi import get_middleware
from starlette.testclient import TestClient

from supertokens_python.recipe import session, thirdparty
from supertokens_python import init

from tests.utils import (
    setup_function,
    teardown_function,
    start_st,
    st_init_common_args,
)


_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore
_ = start_st  # type:ignore


pytestmark = mark.asyncio


@fixture(scope="function")
async def fastapi_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    return TestClient(app)


async def test_thirdpary_parsing_works(fastapi_client: TestClient):
    st_init_args = {
        **st_init_common_args,
        "recipe_list": [
            session.init(),
            thirdparty.init(
                sign_in_and_up_feature=thirdparty.SignInAndUpFeature(
                    providers=[
                        thirdparty.Apple(
                            client_id="4398792-io.supertokens.example.service",
                            client_key_id="7M48Y4RYDL",
                            client_team_id="YWQCXGJRJL",
                            client_private_key="-----BEGIN PRIVATE KEY-----\nMIGTAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBHkwdwIBAQQgu8gXs+XYkqXD6Ala9Sf/iJXzhbwcoG5dMh1OonpdJUmgCgYIKoZIzj0DAQehRANCAASfrvlFbFCYqn3I2zeknYXLwtH30JuOKestDbSfZYxZNMqhF/OzdZFTV0zc5u5s3eN+oCWbnvl0hM+9IW0UlkdA\n-----END PRIVATE KEY-----",
                        )
                    ]
                )
            ),
        ],
    }
    init(**st_init_args)  # type: ignore
    start_st()

    data = {
        "state": "afc596274293e1587315c",
        "code": "c7685e261f98e4b3b94e34b3a69ff9cf4.0.rvxt.eE8rO__6hGoqaX1B7ODPmA",
    }

    res = fastapi_client.post("/auth/callback/apple", data=data)

    assert res.status_code == 200
    assert (
        res.content
        == b'<html><head><script>window.location.replace("http://supertokens.io/auth/callback/apple?state=afc596274293e1587315c&code=c7685e261f98e4b3b94e34b3a69ff9cf4.0.rvxt.eE8rO__6hGoqaX1B7ODPmA");</script></head></html>'
    )



import asyncio
from typing import List, Optional

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.session.asyncio import (
    get_all_session_handles_for_user,
    get_session_information,
)

from ...interfaces import (
    APIInterface,
    APIOptions,
    SessionInfo,
    UserSessionsGetAPIResponse,
)


async def handle_sessions_get(
    _api_interface: APIInterface, api_options: APIOptions
) -> UserSessionsGetAPIResponse:
    user_id = api_options.request.get_query_param("userId")

    if user_id is None:
        raise_bad_input_exception("Missing required parameter 'userId'")

    session_handles = await get_all_session_handles_for_user(user_id)
    sessions: List[Optional[SessionInfo]] = [None for _ in session_handles]

    async def call_(i: int, session_handle: str):
        try:
            session_response = await get_session_information(session_handle)
            if session_response is not None:
                sessions[i] = SessionInfo(session_response)
        except Exception:
            sessions[i] = None

    session_info_promises = [
        call_(i, handle) for i, handle in enumerate(session_handles)
    ]

    await asyncio.gather(*session_info_promises)

    return UserSessionsGetAPIResponse([s for s in sessions if s is not None])



from typing import Dict, Any, Union, Optional

from supertokens_python.framework.request import BaseRequest
from supertokens_python.recipe import session
from supertokens_python.recipe.session import JWTConfig
from supertokens_python.recipe.session.claims import (
    BooleanClaim,
    SessionClaim,
)
from supertokens_python.recipe.session.interfaces import RecipeInterface
from tests.utils import st_init_common_args

TrueClaim = BooleanClaim("st-true", fetch_value=lambda _, __: True)  # type: ignore
NoneClaim = BooleanClaim("st-none", fetch_value=lambda _, __: None)  # type: ignore


def session_functions_override_with_claim(
    claim: SessionClaim[Any], params: Union[Dict[str, Any], None] = None
):
    if params is None:
        params = {}

    def session_function_override(oi: RecipeInterface) -> RecipeInterface:
        oi_create_new_session = oi.create_new_session

        async def new_create_new_session(
            request: BaseRequest,
            user_id: str,
            access_token_payload: Union[None, Dict[str, Any]],
            session_data: Union[None, Dict[str, Any]],
            user_context: Dict[str, Any],
        ):
            payload_update = await claim.build(user_id, user_context)
            if access_token_payload is None:
                access_token_payload = {}
            access_token_payload = {
                **access_token_payload,
                **payload_update,
                **params,
            }

            return await oi_create_new_session(
                request, user_id, access_token_payload, session_data, user_context
            )

        oi.create_new_session = new_create_new_session
        return oi

    return session_function_override


def get_st_init_args(
    claim: SessionClaim[Any] = TrueClaim, jwt: Optional[JWTConfig] = None
):
    return {
        **st_init_common_args,
        "recipe_list": [
            session.init(
                override=session.InputOverrideConfig(
                    functions=session_functions_override_with_claim(claim),
                ),
                jwt=jwt,
            ),
        ],
    }



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, NoReturn

from supertokens_python.exceptions import SuperTokensError

if TYPE_CHECKING:
    from .types import ErrorFormField


def raise_form_field_exception(msg: str, form_fields: List[ErrorFormField]) -> NoReturn:
    raise FieldError(msg, form_fields)


class SuperTokensEmailPasswordError(SuperTokensError):
    pass


class FieldError(SuperTokensEmailPasswordError):
    def __init__(self, msg: str, form_fields: List[ErrorFormField]):
        super().__init__(msg)
        self.form_fields = form_fields

    def get_json_form_fields(self) -> List[Dict[str, Any]]:
        form_fields: List[Dict[str, Any]] = []
        for form_field in self.form_fields:
            form_fields.append({"id": form_field.id, "error": form_field.error})
        return form_fields



from typing import Any, Dict, Union

from supertokens_python.recipe.usermetadata.recipe import UserMetadataRecipe


async def get_user_metadata(
    user_id: str, user_context: Union[Dict[str, Any], None] = None
):
    if user_context is None:
        user_context = {}
    return (
        await UserMetadataRecipe.get_instance().recipe_implementation.get_user_metadata(
            user_id, user_context
        )
    )


async def update_user_metadata(
    user_id: str,
    metadata_update: Dict[str, Any],
    user_context: Union[Dict[str, Any], None] = None,
):
    if user_context is None:
        user_context = {}
    return await UserMetadataRecipe.get_instance().recipe_implementation.update_user_metadata(
        user_id, metadata_update, user_context
    )


async def clear_user_metadata(
    user_id: str, user_context: Union[Dict[str, Any], None] = None
):
    if user_context is None:
        user_context = {}
    return await UserMetadataRecipe.get_instance().recipe_implementation.clear_user_metadata(
        user_id, user_context
    )



from typing import Union

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.dashboard.utils import get_user_for_recipe_id
from supertokens_python.recipe.usermetadata import UserMetadataRecipe
from supertokens_python.recipe.usermetadata.asyncio import get_user_metadata

from ...interfaces import (
    APIInterface,
    APIOptions,
    UserGetAPINoUserFoundError,
    UserGetAPIOkResponse,
    UserGetAPIRecipeNotInitialisedError,
)
from ...utils import is_recipe_initialised, is_valid_recipe_id


async def handle_user_get(
    _api_interface: APIInterface, api_options: APIOptions
) -> Union[
    UserGetAPINoUserFoundError,
    UserGetAPIOkResponse,
    UserGetAPIRecipeNotInitialisedError,
]:
    user_id = api_options.request.get_query_param("userId")
    recipe_id = api_options.request.get_query_param("recipeId")

    if user_id is None:
        raise_bad_input_exception("Missing required parameter 'userId'")

    if recipe_id is None:
        raise_bad_input_exception("Missing required parameter 'recipeId'")

    if not is_valid_recipe_id(recipe_id):
        raise_bad_input_exception("Invalid recipe id")

    if not is_recipe_initialised(recipe_id):
        return UserGetAPIRecipeNotInitialisedError()

    user_response = await get_user_for_recipe_id(user_id, recipe_id)
    if user_response is None:
        return UserGetAPINoUserFoundError()

    user = user_response.user

    try:
        UserMetadataRecipe.get_instance()
    except Exception:
        user.first_name = "FEATURE_NOT_ENABLED"
        user.last_name = "FEATURE_NOT_ENABLED"

        return UserGetAPIOkResponse(recipe_id, user)

    user_metadata = await get_user_metadata(user_id)
    first_name = user_metadata.metadata.get("first_name", "")
    last_name = user_metadata.metadata.get("last_name", "")

    user.first_name = first_name
    user.last_name = last_name

    return UserGetAPIOkResponse(recipe_id, user)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime, timedelta
from typing import Any, Dict, List
from unittest.mock import MagicMock

from fastapi import FastAPI
from fastapi.requests import Request
from fastapi.testclient import TestClient
from pytest import fixture, mark

from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi.fastapi_middleware import get_middleware
from supertokens_python.process_state import AllowedProcessStates, ProcessState
from supertokens_python.recipe import session
from supertokens_python.recipe.session import InputOverrideConfig, SessionRecipe
from supertokens_python.recipe.session.asyncio import (
    create_new_session as async_create_new_session,
)
from supertokens_python.recipe.session.asyncio import (
    get_all_session_handles_for_user,
    get_session_information,
    regenerate_access_token,
)
from supertokens_python.recipe.session.asyncio import (
    revoke_session as asyncio_revoke_session,
)
from supertokens_python.recipe.session.asyncio import (
    update_access_token_payload,
    update_session_data,
)
from supertokens_python.recipe.session.interfaces import RecipeInterface
from supertokens_python.recipe.session.jwt import (
    parse_jwt_without_signature_verification,
)
from supertokens_python.recipe.session.recipe_implementation import RecipeImplementation
from supertokens_python.recipe.session.session_functions import (
    create_new_session,
    get_session,
    refresh_session,
    revoke_session,
)
from tests.utils import clean_st, reset, setup_st, start_st

pytestmark = mark.asyncio


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


async def test_that_once_the_info_is_loaded_it_doesnt_query_again():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[session.init(anti_csrf="VIA_TOKEN")],
    )
    start_st()

    s = SessionRecipe.get_instance()
    if not isinstance(s.recipe_implementation, RecipeImplementation):
        raise Exception("Should never come here")

    response = await create_new_session(s.recipe_implementation, "", False, {}, {})

    assert response["session"] is not None
    assert response["accessToken"] is not None
    assert response["refreshToken"] is not None
    assert response["idRefreshToken"] is not None
    assert response["antiCsrfToken"] is not None
    assert len(response.keys()) == 5

    access_token = parse_jwt_without_signature_verification(
        response["accessToken"]["token"]
    )

    await get_session(
        s.recipe_implementation,
        access_token,
        response["antiCsrfToken"],
        True,
        response["idRefreshToken"]["token"],
    )
    assert (
        AllowedProcessStates.CALLING_SERVICE_IN_VERIFY
        not in ProcessState.get_instance().history
    )

    response2 = await refresh_session(
        s.recipe_implementation,
        response["refreshToken"]["token"],
        response["antiCsrfToken"],
        True,
        "cookie",
    )

    assert response2["session"] is not None
    assert response2["accessToken"] is not None
    assert response2["refreshToken"] is not None
    assert response2["idRefreshToken"] is not None
    assert response2["antiCsrfToken"] is not None
    assert len(response.keys()) == 5

    access_token2 = parse_jwt_without_signature_verification(
        response2["accessToken"]["token"]
    )

    response3 = await get_session(
        s.recipe_implementation,
        access_token2,
        response2["antiCsrfToken"],
        True,
        response["idRefreshToken"]["token"],
    )

    assert (
        AllowedProcessStates.CALLING_SERVICE_IN_VERIFY
        in ProcessState.get_instance().history
    )

    assert response3["session"] is not None
    assert response3["accessToken"] is not None
    assert len(response3.keys()) == 2

    ProcessState.get_instance().reset()

    access_token3 = parse_jwt_without_signature_verification(
        response3["accessToken"]["token"]
    )

    response4 = await get_session(
        s.recipe_implementation,
        access_token3,
        response2["antiCsrfToken"],
        True,
        response["idRefreshToken"]["token"],
    )
    assert (
        AllowedProcessStates.CALLING_SERVICE_IN_VERIFY
        not in ProcessState.get_instance().history
    )

    assert response4["session"] is not None
    assert "accessToken" not in response4
    assert len(response4.keys()) == 1

    response5 = await revoke_session(
        s.recipe_implementation, response4["session"]["handle"]
    )

    assert response5 is True


async def test_creating_many_sessions_for_one_user_and_looping():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    )
    start_st()

    s = SessionRecipe.get_instance()
    if not isinstance(s.recipe_implementation, RecipeImplementation):
        raise Exception("Should never come here")

    access_tokens: List[str] = []
    for _ in range(7):
        new_session = await create_new_session(
            s.recipe_implementation, "someUser", False, {"someKey": "someValue"}, {}
        )
        access_tokens.append(new_session["accessToken"]["token"])

    session_handles = await get_all_session_handles_for_user("someUser")

    assert len(session_handles) == 7

    for i, handle in enumerate(session_handles):
        info = await get_session_information(handle)
        assert info is not None
        assert info.user_id == "someUser"
        assert info.access_token_payload["someKey"] == "someValue"

        is_updated = await update_access_token_payload(
            handle, {"someKey2": "someValue"}
        )
        assert is_updated

        is_updated = await update_session_data(handle, {"foo": "bar"})
        assert is_updated

    # Confirm that update funcs worked:
    for handle in session_handles:
        info = await get_session_information(handle)
        assert info is not None
        assert info.user_id == "someUser"
        assert info.access_token_payload == {"someKey2": "someValue"}
        assert info.session_data == {"foo": "bar"}

    # Regenerate access token with new access_token_payload
    for i, token in enumerate(access_tokens):
        result = await regenerate_access_token(token, {"bar": "baz"})
        assert result is not None
        assert (
            result.session.handle == session_handles[i]
        )  # Session handle should remain the same

        # Confirm that update worked:
        info = await get_session_information(result.session.handle)
        assert info is not None
        assert info.access_token_payload == {"bar": "baz"}

    # Try updating invalid handles:
    is_updated = await update_access_token_payload("invalidHandle", {"foo": "bar"})
    assert is_updated is False
    is_updated = await update_session_data("invalidHandle", {"foo": "bar"})
    assert is_updated is False


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.get("/")
    async def home(_request: Request):  # type: ignore
        return {"hello": "world"}

    @app.post("/create")
    async def create_api(request: Request):  # type: ignore
        await async_create_new_session(request, "test-user", {}, {})
        return ""

    return TestClient(app)


async def test_signout_api_works_even_if_session_is_deleted_after_creation(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[session.init(anti_csrf="VIA_TOKEN")],
    )
    start_st()

    s = SessionRecipe.get_instance()
    if not isinstance(s.recipe_implementation, RecipeImplementation):
        raise Exception("Should never come here")
    user_id = "user_id"

    response = await create_new_session(s.recipe_implementation, user_id, False, {}, {})

    session_handle = response["session"]["handle"]

    revoked = await asyncio_revoke_session(session_handle)
    assert revoked

    signout_response = driver_config_client.post(
        url="/auth/signout",
        cookies={
            "sAccessToken": response["accessToken"]["token"],
        },
        headers={"anti-csrf": response.get("antiCsrfToken", "")},
    )

    assert signout_response.json() == {"status": "OK"}

    assert (
        signout_response.headers["set-cookie"]
        == """sAccessToken=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; HttpOnly; Path=/; SameSite=lax; Secure, sRefreshToken=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; HttpOnly; Path=/auth/session/refresh; SameSite=lax; Secure"""
    )


async def test_should_use_override_functions_in_session_container_methods():
    def override_session_functions(oi: RecipeInterface) -> RecipeInterface:
        oi_get_session_information = oi.get_session_information

        async def get_session_information(
            session_handle: str, user_context: Dict[str, Any]
        ):
            info = await oi_get_session_information(session_handle, user_context)
            assert info is not None
            info.session_data["foo"] = "bar"
            return info

        oi.get_session_information = get_session_information

        return oi

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="https://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                override=InputOverrideConfig(
                    functions=override_session_functions,
                ),
            )
        ],
    )
    start_st()

    s = SessionRecipe.get_instance()
    if not isinstance(s.recipe_implementation, RecipeImplementation):
        raise Exception("Should never come here")

    mock_response = MagicMock()

    my_session = await async_create_new_session(mock_response, "test_id")
    data = await my_session.get_session_data()

    assert data == {"foo": "bar"}


from supertokens_python.recipe.session.exceptions import raise_unauthorised_exception
from supertokens_python.recipe.session.interfaces import APIInterface, APIOptions
from tests.utils import (
    assert_info_clears_tokens,
    extract_all_cookies,
    extract_info,
    get_st_init_args,
)


async def test_revoking_session_during_refresh_with_revoke_session_with_200(
    driver_config_client: TestClient,
):
    def session_api_override(oi: APIInterface) -> APIInterface:
        oi_refresh_post = oi.refresh_post

        async def refresh_post(api_options: APIOptions, user_context: Dict[str, Any]):
            s = await oi_refresh_post(api_options, user_context)
            await s.revoke_session()
            return s

        oi.refresh_post = refresh_post
        return oi

    init_args = get_st_init_args(
        [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(apis=session_api_override),
            )
        ]
    )
    init(**init_args)
    start_st()

    response = driver_config_client.post("/create")
    cookies = extract_all_cookies(response)

    assert "sAccessToken" in cookies
    assert "sRefreshToken" in cookies

    assert "anti-csrf" in response.headers
    assert "front-token" in response.headers

    response = driver_config_client.post(
        "/auth/session/refresh",
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        headers={"anti-csrf": response.headers["anti-csrf"]},
    )

    assert response.status_code == 200
    info = extract_info(response)
    assert_info_clears_tokens(info, "cookie")


async def test_revoking_session_during_refresh_with_revoke_session_sending_401(
    driver_config_client: TestClient,
):
    def session_api_override(oi: APIInterface) -> APIInterface:
        oi_refresh_post = oi.refresh_post

        async def refresh_post(api_options: APIOptions, user_context: Dict[str, Any]):
            s = await oi_refresh_post(api_options, user_context)
            await s.revoke_session()
            api_options.response.set_status_code(401)  # type: ignore
            api_options.response.set_json_content({})  # type: ignore
            return s

        oi.refresh_post = refresh_post
        return oi

    init_args = get_st_init_args(
        [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(apis=session_api_override),
            )
        ]
    )
    init(**init_args)
    start_st()

    response = driver_config_client.post("/create")
    assert response.status_code == 200

    cookies = extract_all_cookies(response)

    assert "sAccessToken" in cookies
    assert "sRefreshToken" in cookies

    assert "anti-csrf" in response.headers
    assert "front-token" in response.headers

    response = driver_config_client.post(
        "/auth/session/refresh",
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        headers={"anti-csrf": response.headers["anti-csrf"]},
    )

    assert response.status_code == 401
    info = extract_info(response)
    assert_info_clears_tokens(info, "cookie")


async def test_revoking_session_during_refresh_and_throw_unauthorized(
    driver_config_client: TestClient,
):
    def session_api_override(oi: APIInterface) -> APIInterface:
        oi_refresh_post = oi.refresh_post

        async def refresh_post(api_options: APIOptions, user_context: Dict[str, Any]):
            await oi_refresh_post(api_options, user_context)
            return raise_unauthorised_exception("unauthorized", clear_tokens=True)

        oi.refresh_post = refresh_post
        return oi

    init_args = get_st_init_args(
        [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(apis=session_api_override),
            )
        ]
    )
    init(**init_args)
    start_st()

    response = driver_config_client.post("/create")
    assert response.status_code == 200

    cookies = extract_all_cookies(response)

    assert "sAccessToken" in cookies
    assert "sRefreshToken" in cookies

    assert "anti-csrf" in response.headers
    assert "front-token" in response.headers

    response = driver_config_client.post(
        "/auth/session/refresh",
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        headers={"anti-csrf": response.headers["anti-csrf"]},
    )

    assert response.status_code == 401
    cookies = extract_all_cookies(response)

    assert (
        "anti-csrf" not in response.headers
    )  # TODO: This makes sense. But verify this
    assert response.headers["front-token"] != ""

    assert cookies["sAccessToken"]["value"] == ""
    assert cookies["sRefreshToken"]["value"] == ""
    assert cookies["sAccessToken"]["expires"] == "Thu, 01 Jan 1970 00:00:00 GMT"
    assert cookies["sRefreshToken"]["expires"] == "Thu, 01 Jan 1970 00:00:00 GMT"

    assert cookies["sAccessToken"]["domain"] == ""
    assert cookies["sRefreshToken"]["domain"] == ""


async def test_revoking_session_during_refresh_fails_if_just_sending_401(
    driver_config_client: TestClient,
):
    def session_api_override(oi: APIInterface) -> APIInterface:
        oi_refresh_post = oi.refresh_post

        async def refresh_post(api_options: APIOptions, user_context: Dict[str, Any]):
            s = await oi_refresh_post(api_options, user_context)
            api_options.response.set_status_code(401)  # type: ignore
            api_options.response.set_json_content({})  # type: ignore
            return s

        oi.refresh_post = refresh_post
        return oi

    init_args = get_st_init_args(
        [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
                override=session.InputOverrideConfig(apis=session_api_override),
            )
        ]
    )
    init(**init_args)
    start_st()

    response = driver_config_client.post("/create")
    assert response.status_code == 200

    cookies = extract_all_cookies(response)

    assert "sAccessToken" in cookies
    assert "sRefreshToken" in cookies

    assert "anti-csrf" in response.headers
    assert "front-token" in response.headers

    response = driver_config_client.post(
        "/auth/session/refresh",
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        headers={"anti-csrf": response.headers["anti-csrf"]},
    )

    assert response.status_code == 401
    cookies = extract_all_cookies(response)

    assert response.headers["anti-csrf"] != ""
    assert response.headers["front-token"] != ""

    assert cookies["sAccessToken"]["value"] != ""
    assert cookies["sRefreshToken"]["value"] != ""


async def test_token_cookie_expires(
    driver_config_client: TestClient,
):
    init_args = get_st_init_args(
        [
            session.init(
                anti_csrf="VIA_TOKEN",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
        ]
    )
    init(**init_args)
    start_st()

    response = driver_config_client.post("/create")
    assert response.status_code == 200

    cookies = extract_all_cookies(response)

    assert "sAccessToken" in cookies
    assert "sRefreshToken" in cookies

    for c in response.cookies:
        if c.name == "sAccessToken":  # 100 years (set by the SDK)
            # some time must have elasped since the cookie was set. So less than current time
            assert (
                datetime.fromtimestamp(c.expires or 0) - timedelta(days=365.25 * 100)
                < datetime.now()
            )
        if c.name == "sRefreshToken":  # 100 days (set by the core)
            assert (
                datetime.fromtimestamp(c.expires or 0) - timedelta(days=100)
                < datetime.now()
            )

    assert response.headers["anti-csrf"] != ""
    assert response.headers["front-token"] != ""

    response = driver_config_client.post(
        "/auth/session/refresh",
        cookies={
            "sRefreshToken": cookies["sRefreshToken"]["value"],
        },
        headers={"anti-csrf": response.headers["anti-csrf"]},
    )

    assert response.status_code == 200
    cookies = extract_all_cookies(response)

    assert "sAccessToken" in cookies
    assert "sRefreshToken" in cookies

    for c in response.cookies:
        if c.name == "sAccessToken":  # 100 years (set by the SDK)
            # some time must have elasped since the cookie was set. So less than current time
            assert (
                datetime.fromtimestamp(c.expires or 0) - timedelta(days=365.25 * 100)
                < datetime.now()
            )
        if c.name == "sRefreshToken":  # 100 days (set by the core)
            assert (
                datetime.fromtimestamp(c.expires or 0) - timedelta(days=100)
                < datetime.now()
            )

    assert response.headers["anti-csrf"] != ""
    assert response.headers["front-token"] != ""



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from pytest import fixture, mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import emailpassword, session

from fastapi import FastAPI
from fastapi.testclient import TestClient
from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())
    return TestClient(app)


@mark.asyncio
async def test_rid_with_session_and_non_existent_api_in_session_recipe_gives_404(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(),
        ],
    )
    start_st()

    response = driver_config_client.post(url="/auth/signin", headers={"rid": "session"})
    assert response.status_code == 404


@mark.asyncio
async def test_no_rid_with_existent_API_does_not_give_404(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(),
        ],
    )
    start_st()

    response = driver_config_client.post(url="/auth/signin")
    assert response.status_code == 400


@mark.asyncio
async def test_rid_as_anticsrf_with_existent_API_does_not_give_404(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(),
        ],
    )
    start_st()

    response = driver_config_client.post(
        url="/auth/signin", headers={"rid": "anti-csrf"}
    )
    assert response.status_code == 400


@mark.asyncio
async def test_random_rid_with_existent_API_does_gives_404(
    driver_config_client: TestClient,
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
            emailpassword.init(),
        ],
    )
    start_st()

    response = driver_config_client.post(url="/auth/signin", headers={"rid": "random"})
    assert response.status_code == 404



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Union, Optional

from jwt import decode

from supertokens_python.utils import get_timestamp_ms

from .constants import ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY
from .session_class import get_session_with_jwt
from .utills import add_jwt_to_access_token_payload

if TYPE_CHECKING:
    from supertokens_python.recipe.session.utils import SessionConfig
    from supertokens_python.recipe.session.interfaces import (
        RecipeInterface,
        SessionContainer,
        SessionInformationResult,
    )
    from supertokens_python.framework.types import BaseRequest

from math import ceil

from supertokens_python.recipe.openid.interfaces import (
    RecipeInterface as OpenIdRecipeInterface,
)

EXPIRY_OFFSET_SECONDS = 30


def get_jwt_expiry(access_token_expiry: int):
    return access_token_expiry + EXPIRY_OFFSET_SECONDS


def get_recipe_implementation_with_jwt(
    original_implementation: RecipeInterface,
    config: SessionConfig,
    openid_recipe_implementation: OpenIdRecipeInterface,
) -> RecipeInterface:

    og_create_new_session = original_implementation.create_new_session

    async def create_new_session(
        request: BaseRequest,
        user_id: str,
        access_token_payload: Union[None, Dict[str, Any]],
        session_data: Union[None, Dict[str, Any]],
        user_context: Dict[str, Any],
    ) -> SessionContainer:
        if access_token_payload is None:
            access_token_payload = {}
        access_token_validity_in_seconds = ceil(
            await original_implementation.get_access_token_lifetime_ms(user_context)
            / 1000
        )
        access_token_payload = await add_jwt_to_access_token_payload(
            access_token_payload=access_token_payload,
            jwt_expiry=get_jwt_expiry(access_token_validity_in_seconds),
            user_id=user_id,
            jwt_property_name=config.jwt.property_name_in_access_token_payload,
            openid_recipe_implementation=openid_recipe_implementation,
            user_context=user_context,
        )
        session = await og_create_new_session(
            request,
            user_id,
            access_token_payload,
            session_data,
            user_context=user_context,
        )
        return get_session_with_jwt(session, openid_recipe_implementation)

    og_get_session = original_implementation.get_session

    async def get_session(
        request: BaseRequest,
        anti_csrf_check: Union[bool, None],
        session_required: bool,
        user_context: Dict[str, Any],
    ) -> Union[SessionContainer, None]:
        session_container = await og_get_session(
            request,
            anti_csrf_check,
            session_required,
            user_context,
        )
        if session_container is None:
            return None
        return get_session_with_jwt(session_container, openid_recipe_implementation)

    og_refresh_session = original_implementation.refresh_session

    async def refresh_session(
        request: BaseRequest, user_context: Dict[str, Any]
    ) -> SessionContainer:
        access_token_validity_in_seconds = ceil(
            await original_implementation.get_access_token_lifetime_ms(user_context)
            / 1000
        )

        # Refresh session first because this will create a new access token
        new_session = await og_refresh_session(request, user_context)
        access_token_payload = new_session.get_access_token_payload()
        access_token_payload = await add_jwt_to_access_token_payload(
            access_token_payload=access_token_payload,
            jwt_expiry=get_jwt_expiry(access_token_validity_in_seconds),
            user_id=new_session.get_user_id(),
            jwt_property_name=config.jwt.property_name_in_access_token_payload,
            openid_recipe_implementation=openid_recipe_implementation,
            user_context=user_context,
        )

        await new_session.update_access_token_payload(
            access_token_payload, user_context
        )
        return get_session_with_jwt(new_session, openid_recipe_implementation)

    og_update_access_token_payload = original_implementation.update_access_token_payload

    async def jwt_aware_update_access_token_payload(
        session_information: SessionInformationResult,
        new_access_token_payload: Dict[str, Any],
        user_context: Dict[str, Any],
    ) -> bool:
        access_token_payload = session_information.access_token_payload

        if ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY not in access_token_payload:
            return await og_update_access_token_payload(
                session_information.session_handle,
                new_access_token_payload,
                user_context,
            )

        existing_jwt_property_name = access_token_payload[
            ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY
        ]

        assert existing_jwt_property_name in access_token_payload
        existing_jwt = access_token_payload[existing_jwt_property_name]

        current_time_in_seconds = ceil(get_timestamp_ms() / 1000)
        decoded_payload: Dict[str, Any] = decode(
            jwt=existing_jwt, options={"verify_signature": False, "verify_exp": False}
        )

        if decoded_payload is None or decoded_payload.get("exp") is None:
            raise Exception("Error reading JWT from session")

        jwt_expiry = 1
        if "exp" in decoded_payload:
            exp = decoded_payload["exp"]
            if exp > current_time_in_seconds:
                # it can come here if someone calls this function well after
                # the access token and the jwt payload have expired. In this case,
                # we still want the jwt payload to update, but the resulting JWT should
                # not be alive for too long (since it's expired already). So we set it to
                # 1 second lifetime.
                jwt_expiry = exp - current_time_in_seconds

        new_access_token_payload = await add_jwt_to_access_token_payload(
            access_token_payload=new_access_token_payload,
            jwt_expiry=jwt_expiry,
            user_id=session_information.user_id,
            jwt_property_name=existing_jwt_property_name,
            openid_recipe_implementation=openid_recipe_implementation,
            user_context=user_context,
        )

        return await og_update_access_token_payload(
            session_information.session_handle, new_access_token_payload, user_context
        )

    async def update_access_token_payload(
        session_handle: str,
        new_access_token_payload: Optional[Dict[str, Any]],
        user_context: Dict[str, Any],
    ) -> bool:
        if new_access_token_payload is None:
            new_access_token_payload = {}

        session_information = await original_implementation.get_session_information(
            session_handle, user_context
        )
        if session_information is None:
            return False

        return await jwt_aware_update_access_token_payload(
            session_information, new_access_token_payload, user_context
        )

    async def merge_into_access_token_payload(
        session_handle: str,
        access_token_payload_update: Dict[str, Any],
        user_context: Dict[str, Any],
    ) -> bool:
        session_information = await original_implementation.get_session_information(
            session_handle, user_context
        )
        if session_information is None:
            return False

        new_access_token_payload = {
            **session_information.access_token_payload,
            **access_token_payload_update,
        }
        for k in access_token_payload_update.keys():
            if new_access_token_payload[k] is None:
                del new_access_token_payload[k]

        return await jwt_aware_update_access_token_payload(
            session_information, new_access_token_payload, user_context
        )

    original_implementation.create_new_session = create_new_session
    original_implementation.get_session = get_session
    original_implementation.refresh_session = refresh_session
    original_implementation.update_access_token_payload = update_access_token_payload
    original_implementation.merge_into_access_token_payload = (
        merge_into_access_token_payload
    )
    return original_implementation



from typing import Any, Dict, Union

from supertokens_python.async_to_sync_wrapper import sync


def get_user_metadata(user_id: str, user_context: Union[Dict[str, Any], None] = None):
    from supertokens_python.recipe.usermetadata.asyncio import get_user_metadata

    return sync(get_user_metadata(user_id, user_context))


def update_user_metadata(
    user_id: str,
    metadata_update: Dict[str, Any],
    user_context: Union[Dict[str, Any], None] = None,
):
    from supertokens_python.recipe.usermetadata.asyncio import update_user_metadata

    return sync(update_user_metadata(user_id, metadata_update, user_context))


def clear_user_metadata(user_id: str, user_context: Union[Dict[str, Any], None] = None):
    from supertokens_python.recipe.usermetadata.asyncio import clear_user_metadata

    return sync(clear_user_metadata(user_id, user_context))



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Callable, Dict, Optional

from supertokens_python.types import MaybeAwaitable

from .primitive_claim import PrimitiveClaim, PrimitiveClaimValidators


class BooleanClaimValidators(PrimitiveClaimValidators[bool]):
    def is_true(self, max_age: Optional[int], id_: Optional[str] = None):
        return self.has_value(True, max_age, id_)

    def is_false(self, max_age: Optional[int], id_: Optional[str] = None):
        return self.has_value(False, max_age, id_)


class BooleanClaim(PrimitiveClaim[bool]):
    def __init__(
        self,
        key: str,
        fetch_value: Callable[
            [str, Dict[str, Any]],
            MaybeAwaitable[Optional[bool]],
        ],
        default_max_age_in_sec: Optional[int] = None,
    ):
        super().__init__(key, fetch_value, default_max_age_in_sec)
        self.validators = BooleanClaimValidators(
            claim=self, default_max_age_in_sec=default_max_age_in_sec
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import TypeVar

from supertokens_python.ingredients.smsdelivery.types import TwilioSettings

_T = TypeVar("_T")


def normalize_twilio_settings(twilio_settings: TwilioSettings) -> TwilioSettings:
    from_ = twilio_settings.from_
    messaging_service_sid = twilio_settings.messaging_service_sid

    if (from_ and messaging_service_sid) or (not from_ and not messaging_service_sid):
        raise Exception(
            'Please pass exactly one of "from" and "messaging_service_sid" config for twilio_settings.'
        )

    return twilio_settings



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import List

from pytest import mark
from tests.utils import clean_st, reset, setup_st, start_st

from supertokens_python import InputAppInfo, SupertokensConfig
from supertokens_python import asyncio as st_asyncio
from supertokens_python import init
from supertokens_python.querier import Querier
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword import asyncio as ep_asyncio
from supertokens_python.recipe.emailpassword.interfaces import SignUpOkResult
from supertokens_python.utils import is_version_gte


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@mark.asyncio
async def test_supertokens_functions():
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="ST",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(),
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie"),
        ],
    )
    start_st()

    emails = [f"{u}@example.com" for u in ["foo", "bar", "baz"]]
    user_ids: List[str] = []
    for e in emails:
        signup_resp = await ep_asyncio.sign_up(e, "secret_pass")
        assert isinstance(signup_resp, SignUpOkResult)
        user_ids.append(signup_resp.user.user_id)

    # Get user count
    assert await st_asyncio.get_user_count() == len(emails)

    # Get users in ascending order by joining time
    users_asc = (await st_asyncio.get_users_oldest_first(limit=10)).users
    emails_asc = [user.email for user in users_asc]
    assert emails_asc == emails

    # Get users in descending order by joining time
    users_desc = (await st_asyncio.get_users_newest_first(limit=10)).users
    emails_desc = [user.email for user in users_desc]
    assert emails_desc == emails[::-1]

    version = await Querier.get_instance().get_api_version()
    if not is_version_gte(version, "2.10"):
        # If the version is less than 2.10, delete user feature doesn't exist, so mark the test successful
        return

    # Delete the 2nd user (bar@example.com)
    await st_asyncio.delete_user(user_ids[1])

    # Again, get users in ascending order by joining time
    # We expect that the 2nd user (bar@example.com) must be absent.
    users_asc = (await st_asyncio.get_users_oldest_first(limit=10)).users
    emails_asc = [user.email for user in users_asc]
    assert emails[1] not in emails_asc  # The 2nd user must be deleted now.



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Union

from httpx import AsyncClient

from supertokens_python.ingredients.emaildelivery.types import EmailDeliveryInterface
from supertokens_python.logger import log_debug_message
from supertokens_python.recipe.emailpassword.interfaces import (
    EmailTemplateVars,
    RecipeInterface,
)
from supertokens_python.recipe.emailpassword.types import User
from supertokens_python.supertokens import AppInfo
from supertokens_python.utils import handle_httpx_client_exceptions

if TYPE_CHECKING:
    from supertokens_python.recipe.emailpassword.utils import (
        InputResetPasswordUsingTokenFeature,
    )


def default_create_and_send_custom_email(
    app_info: AppInfo,
) -> Callable[[User, str, Dict[str, Any]], Awaitable[None]]:
    async def func(user: User, password_reset_url_with_token: str, _: Dict[str, Any]):
        if ("SUPERTOKENS_ENV" in environ) and (environ["SUPERTOKENS_ENV"] == "testing"):
            return
        data = {
            "email": user.email,
            "appName": app_info.app_name,
            "passwordResetURL": password_reset_url_with_token,
        }
        try:
            async with AsyncClient() as client:
                resp = await client.post("https://api.supertokens.io/0/st/auth/password/reset", json=data, headers={"api-version": "0"})  # type: ignore
                resp.raise_for_status()
                log_debug_message("Password reset email sent to %s", user.email)
        except Exception as e:
            log_debug_message("Error sending password reset email")
            handle_httpx_client_exceptions(e, data)

    return func


class BackwardCompatibilityService(EmailDeliveryInterface[EmailTemplateVars]):
    app_info: AppInfo

    def __init__(
        self,
        app_info: AppInfo,
        recipe_interface_impl: RecipeInterface,
        reset_password_using_token_feature: Union[
            InputResetPasswordUsingTokenFeature, None
        ] = None,
    ) -> None:
        self.recipe_interface_impl = recipe_interface_impl

        reset_password_feature_send_email_func = default_create_and_send_custom_email(
            app_info
        )
        if (
            reset_password_using_token_feature
            and reset_password_using_token_feature.create_and_send_custom_email
            is not None
        ):
            reset_password_feature_send_email_func = (
                reset_password_using_token_feature.create_and_send_custom_email
            )

        self.reset_password_feature_send_email_func = (
            reset_password_feature_send_email_func
        )

    async def send_email(
        self, template_vars: EmailTemplateVars, user_context: Dict[str, Any]
    ) -> None:
        user = await self.recipe_interface_impl.get_user_by_id(
            user_id=template_vars.user.id, user_context=user_context
        )
        if user is None:
            raise Exception("Should never come here")

        # we add this here cause the user may have overridden the sendEmail function
        # to change the input email and if we don't do this, the input email
        # will get reset by the getUserById call above.
        user.email = template_vars.user.email
        try:
            await self.reset_password_feature_send_email_func(
                user, template_vars.password_reset_link, user_context
            )
        except Exception:
            pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Union

from supertokens_python.recipe.thirdpartyemailpassword.recipe import (
    ThirdPartyEmailPasswordRecipe,
)

from ..types import EmailTemplateVars, User


async def get_user_by_id(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[None, User]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.get_user_by_id(
        user_id, user_context
    )


async def get_user_by_third_party_info(
    third_party_id: str,
    third_party_user_id: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.get_user_by_thirdparty_info(
        third_party_id, third_party_user_id, user_context
    )


async def thirdparty_sign_in_up(
    third_party_id: str,
    third_party_user_id: str,
    email: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.thirdparty_sign_in_up(
        third_party_id, third_party_user_id, email, user_context
    )


async def create_reset_password_token(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.create_reset_password_token(
        user_id, user_context
    )


async def reset_password_using_token(
    token: str, new_password: str, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.reset_password_using_token(
        token, new_password, user_context
    )


async def emailpassword_sign_in(
    email: str, password: str, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.emailpassword_sign_in(
        email, password, user_context
    )


async def emailpassword_sign_up(
    email: str, password: str, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.emailpassword_sign_up(
        email, password, user_context
    )


async def update_email_or_password(
    user_id: str,
    email: Union[None, str] = None,
    password: Union[None, str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.update_email_or_password(
        user_id, email, password, user_context
    )


async def get_users_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[User]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().recipe_implementation.get_users_by_email(
        email, user_context
    )


async def send_email(
    input_: EmailTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyEmailPasswordRecipe.get_instance().email_delivery.ingredient_interface_impl.send_email(
        input_, user_context
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Union

if TYPE_CHECKING:
    from supertokens_python.recipe.session.interfaces import SessionContainer


class BaseRequest(ABC):
    def __init__(self):
        self.wrapper_used = True
        self.request = None

    @abstractmethod
    def get_query_param(
        self, key: str, default: Union[str, None] = None
    ) -> Union[str, None]:
        pass

    @abstractmethod
    async def json(self) -> Union[Any, None]:
        pass

    @abstractmethod
    async def form_data(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def method(self) -> str:
        pass

    @abstractmethod
    def get_cookie(self, key: str) -> Union[str, None]:
        pass

    @abstractmethod
    def get_header(self, key: str) -> Union[None, str]:
        pass

    @abstractmethod
    def get_session(self) -> Union[SessionContainer, None]:
        pass

    @abstractmethod
    def set_session(self, session: SessionContainer):
        pass

    @abstractmethod
    def set_session_as_none(self):
        """
        This function is used to set the request's session variable to None.
        See https://github.com/supertokens/supertokens-python/issues/90
        """

    @abstractmethod
    def get_path(self) -> str:
        pass



import json
from typing import Any, Dict, TypeVar, Union
from unittest.mock import patch

from pytest import mark
from supertokens_python.recipe import session

from supertokens_python.recipe.session import SessionRecipe
from supertokens_python.recipe.session.recipe_implementation import RecipeImplementation
from supertokens_python.recipe.session.claims import PrimitiveClaim
from supertokens_python.recipe.session.interfaces import (
    JSONObject,
    SessionClaimValidator,
    ClaimValidationResult,
    SessionClaim,
)
from supertokens_python.recipe.session.session_class import Session
from supertokens_python import init
from tests.utils import setup_function, teardown_function, start_st, st_init_common_args

_ = setup_function  # type:ignore
_ = teardown_function  # type:ignore

_T = TypeVar("_T")

pytestmark = mark.asyncio


async def test_should_not_throw_for_empty_array():
    st_args = {
        **st_init_common_args,
        "recipe_list": [
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    }
    init(**st_args)  # type:ignore
    start_st()

    s = SessionRecipe.get_instance()
    assert isinstance(s.recipe_implementation, RecipeImplementation)

    user_session = Session(
        s.recipe_implementation,
        s.recipe_implementation.config,
        "test_access_token",
        "test_session_handle",
        "test_user_id",
        {},
        "cookie",
    )
    with patch.object(
        Session,
        "update_access_token_payload",
        wraps=user_session.update_access_token_payload,
    ) as mock:
        await user_session.assert_claims([])
        mock.assert_not_called()


async def test_should_call_validate_with_the_same_payload_object():
    st_args = {
        **st_init_common_args,
        "recipe_list": [
            session.init(get_token_transfer_method=lambda _, __, ___: "cookie")
        ],
    }
    init(**st_args)  # type:ignore
    start_st()

    s = SessionRecipe.get_instance()
    assert isinstance(s.recipe_implementation, RecipeImplementation)

    payload = {"custom-key": "custom-value"}
    user_session = Session(
        s.recipe_implementation,
        s.recipe_implementation.config,
        "test_access_token",
        "test_session_handle",
        "test_user_id",
        payload,
        "cookie",
    )

    class DummyClaimValidator(SessionClaimValidator):
        def __init__(self, claim: SessionClaim[Any]):
            super().__init__("claim_validator_id")
            self.claim = claim
            self.validate_calls: Dict[str, int] = {}

        async def validate(
            self, payload: JSONObject, user_context: Union[Dict[str, Any], None] = None
        ):
            payload_json = json.dumps(payload)
            self.validate_calls[payload_json] = (
                self.validate_calls.get(payload_json, 0) + 1
            )
            return ClaimValidationResult(is_valid=True)

        def should_refetch(self, payload: JSONObject, user_context: Dict[str, Any]):
            return False

    dummy_claim = PrimitiveClaim("st-claim", lambda _, __: "Hello world")

    dummy_claim_validator = DummyClaimValidator(dummy_claim)

    dummy_claim.validators.dummy_claim_validator = dummy_claim_validator  # type: ignore

    with patch.object(
        Session,
        "update_access_token_payload",
        wraps=user_session.update_access_token_payload,
    ) as mock:
        await user_session.assert_claims([dummy_claim.validators.dummy_claim_validator])  # type: ignore

        assert dummy_claim_validator.validate_calls == {json.dumps(payload): 1}
        mock.assert_not_called()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from typing import Any, Dict, Union

from _pytest.fixtures import fixture
from flask import Flask, g, jsonify, make_response, request
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.flask import Middleware
from supertokens_python.recipe import emailpassword, session, thirdparty
from supertokens_python.recipe.emailpassword.interfaces import APIInterface, APIOptions
from supertokens_python.recipe.session import SessionContainer
from supertokens_python.recipe.session.framework.flask import verify_session
from supertokens_python.recipe.session.syncio import (
    create_new_session,
    get_session,
    refresh_session,
    revoke_session,
)
from tests.Flask.utils import extract_all_cookies
from tests.utils import (
    TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY,
    TEST_ACCESS_TOKEN_MAX_AGE_VALUE,
    TEST_ACCESS_TOKEN_PATH_CONFIG_KEY,
    TEST_ACCESS_TOKEN_PATH_VALUE,
    TEST_COOKIE_DOMAIN_CONFIG_KEY,
    TEST_COOKIE_DOMAIN_VALUE,
    TEST_COOKIE_SAME_SITE_CONFIG_KEY,
    TEST_COOKIE_SECURE_CONFIG_KEY,
    TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH,
    TEST_DRIVER_CONFIG_COOKIE_DOMAIN,
    TEST_DRIVER_CONFIG_COOKIE_SAME_SITE,
    TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH,
    TEST_REFRESH_TOKEN_MAX_AGE_CONFIG_KEY,
    TEST_REFRESH_TOKEN_MAX_AGE_VALUE,
    TEST_REFRESH_TOKEN_PATH_CONFIG_KEY,
    TEST_REFRESH_TOKEN_PATH_KEY_VALUE,
    clean_st,
    reset,
    set_key_value_in_config,
    setup_st,
    start_st,
)


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
def driver_config_app():
    def override_email_password_apis(original_implementation: APIInterface):

        original_func = original_implementation.email_exists_get

        async def email_exists_get(
            email: str, api_options: APIOptions, user_context: Dict[str, Any]
        ):
            response_dict = {"custom": True}
            api_options.response.set_status_code(203)
            api_options.response.set_json_content(response_dict)
            return await original_func(email, api_options, user_context)

        original_implementation.email_exists_get = email_exists_get
        return original_implementation

    app = Flask(__name__)
    app.app_context().push()
    Middleware(app)

    app.testing = True
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
            api_base_path="/auth",
        ),
        framework="flask",
        recipe_list=[
            session.init(
                anti_csrf="VIA_TOKEN",
                cookie_domain="supertokens.io",
                get_token_transfer_method=lambda _, __, ___: "cookie",
            ),
            emailpassword.init(
                override=emailpassword.InputOverrideConfig(
                    apis=override_email_password_apis
                )
            ),
            thirdparty.init(
                sign_in_and_up_feature=thirdparty.SignInAndUpFeature(
                    providers=[
                        thirdparty.Apple(
                            client_id="4398792-io.supertokens.example.service",
                            client_key_id="7M48Y4RYDL",
                            client_team_id="YWQCXGJRJL",
                            client_private_key="-----BEGIN PRIVATE KEY-----\nMIGTAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBHkwdwIBAQQgu8gXs+XYkqXD6Ala9Sf/iJXzhbwcoG5dMh1OonpdJUmgCgYIKoZIzj0DAQehRANCAASfrvlFbFCYqn3I2zeknYXLwtH30JuOKestDbSfZYxZNMqhF/OzdZFTV0zc5u5s3eN+oCWbnvl0hM+9IW0UlkdA\n-----END PRIVATE KEY-----",
                        )
                    ]
                )
            ),
        ],
    )

    @app.route("/test")  # type: ignore
    def t():  # type: ignore
        return jsonify({})

    @app.route("/login")  # type: ignore
    def login():  # type: ignore
        user_id = "userId"
        create_new_session(request, user_id, {}, {})

        return jsonify({"userId": user_id, "session": "ssss"})

    @app.route("/refresh", methods=["POST"])  # type: ignore
    def custom_refresh():  # type: ignore
        response = make_response(jsonify({}))
        refresh_session(request)
        return response

    @app.route("/info", methods=["GET", "OPTIONS"])  # type: ignore
    def custom_info():  # type: ignore
        if request.method == "OPTIONS":  # type: ignore
            return jsonify({"method": "option"})
        response = make_response(jsonify({}))
        get_session(request, True)
        return response

    @app.route("/handle", methods=["GET", "OPTIONS"])  # type: ignore
    def custom_handle_api():  # type: ignore
        if request.method == "OPTIONS":  # type: ignore
            return jsonify({"method": "option"})
        session: Union[None, SessionContainer] = get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        return jsonify({"s": session.get_user_id()})

    @app.route("/handle-session-optional", methods=["GET", "OPTIONS"])  # type: ignore
    @verify_session(session_required=False)
    def optional_session():  # type: ignore
        if request.method == "OPTIONS":  # type: ignore
            return jsonify({"method": "option"})
        session: Union[SessionContainer, None] = g.supertokens  # type: ignore
        if session is None:
            return jsonify({"s": "empty session"})
        return jsonify({"s": session.get_user_id()})

    @app.route("/logout", methods=["POST"])  # type: ignore
    @verify_session(session_required=False)
    def custom_logout():  # type: ignore
        response = make_response(jsonify({}))
        session: Union[None, SessionContainer] = get_session(request, True)
        if session is None:
            raise Exception("Should never come here")
        revoke_session(session.get_user_id())
        return response

    return app


def test_cookie_login_and_refresh(driver_config_app: Any):
    start_st()

    set_key_value_in_config(TEST_COOKIE_SAME_SITE_CONFIG_KEY, "None")
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY, TEST_ACCESS_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_PATH_CONFIG_KEY, TEST_ACCESS_TOKEN_PATH_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_DOMAIN_CONFIG_KEY, TEST_COOKIE_DOMAIN_VALUE)
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_MAX_AGE_CONFIG_KEY, TEST_REFRESH_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_PATH_CONFIG_KEY, TEST_REFRESH_TOKEN_PATH_KEY_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_SECURE_CONFIG_KEY, "false")

    response_1 = driver_config_app.test_client().get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )

    test_client = driver_config_app.test_client()
    test_client.set_cookie(
        "localhost", "sRefreshToken", cookies_1["sRefreshToken"]["value"]
    )
    response_2 = test_client.post(
        "/refresh", headers={"anti-csrf": response_1.headers.get("anti-csrf")}
    )
    cookies_2 = extract_all_cookies(response_2)
    assert cookies_1["sAccessToken"]["value"] != cookies_2["sAccessToken"]["value"]
    assert cookies_1["sRefreshToken"]["value"] != cookies_2["sRefreshToken"]["value"]
    assert response_2.headers.get("anti-csrf") is not None
    assert cookies_2["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_2["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_2["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_2["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_2["sAccessToken"]["httponly"]
    assert cookies_2["sRefreshToken"]["httponly"]
    assert (
        cookies_2["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_2["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )


def test_login_refresh_no_csrf(driver_config_app: Any):
    start_st()

    set_key_value_in_config(TEST_COOKIE_SAME_SITE_CONFIG_KEY, "None")
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY, TEST_ACCESS_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_PATH_CONFIG_KEY, TEST_ACCESS_TOKEN_PATH_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_DOMAIN_CONFIG_KEY, TEST_COOKIE_DOMAIN_VALUE)
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_MAX_AGE_CONFIG_KEY, TEST_REFRESH_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_PATH_CONFIG_KEY, TEST_REFRESH_TOKEN_PATH_KEY_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_SECURE_CONFIG_KEY, "false")

    response_1 = driver_config_app.test_client().get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )

    test_client = driver_config_app.test_client()
    test_client.set_cookie(
        "localhost", "sRefreshToken", cookies_1["sRefreshToken"]["value"]
    )

    # post with csrf token -> no error
    result = test_client.post(
        "/refresh", headers={"anti-csrf": response_1.headers.get("anti-csrf")}
    )
    assert result.status_code == 200

    # post with csrf token -> should be error with status code 401
    result = test_client.post("/refresh")
    assert result.status_code == 401


def test_login_logout(driver_config_app: Any):
    start_st()

    set_key_value_in_config(TEST_COOKIE_SAME_SITE_CONFIG_KEY, "None")
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY, TEST_ACCESS_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_PATH_CONFIG_KEY, TEST_ACCESS_TOKEN_PATH_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_DOMAIN_CONFIG_KEY, TEST_COOKIE_DOMAIN_VALUE)
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_MAX_AGE_CONFIG_KEY, TEST_REFRESH_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_PATH_CONFIG_KEY, TEST_REFRESH_TOKEN_PATH_KEY_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_SECURE_CONFIG_KEY, "false")

    response_1 = driver_config_app.test_client().get("/login")
    cookies_1 = extract_all_cookies(response_1)

    assert response_1.headers.get("anti-csrf") is not None
    assert cookies_1["sAccessToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sRefreshToken"]["domain"] == TEST_DRIVER_CONFIG_COOKIE_DOMAIN
    assert cookies_1["sAccessToken"]["path"] == TEST_DRIVER_CONFIG_ACCESS_TOKEN_PATH
    assert cookies_1["sRefreshToken"]["path"] == TEST_DRIVER_CONFIG_REFRESH_TOKEN_PATH
    assert cookies_1["sAccessToken"]["httponly"]
    assert cookies_1["sRefreshToken"]["httponly"]
    assert (
        cookies_1["sAccessToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )
    assert (
        cookies_1["sRefreshToken"]["samesite"].lower()
        == TEST_DRIVER_CONFIG_COOKIE_SAME_SITE
    )

    test_client = driver_config_app.test_client()
    test_client.set_cookie(
        "localhost", "sAccessToken", cookies_1["sAccessToken"]["value"]
    )

    response_2 = test_client.post(
        "/logout", headers={"anti-csrf": response_1.headers.get("anti-csrf")}
    )

    cookies_2 = extract_all_cookies(response_2)
    assert not cookies_2

    response_3 = test_client.post(
        "/logout", headers={"anti-csrf": response_1.headers.get("anti-csrf")}
    )

    assert response_3.status_code == 200


def test_login_handle(driver_config_app: Any):
    start_st()

    set_key_value_in_config(TEST_COOKIE_SAME_SITE_CONFIG_KEY, "None")
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_MAX_AGE_CONFIG_KEY, TEST_ACCESS_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_ACCESS_TOKEN_PATH_CONFIG_KEY, TEST_ACCESS_TOKEN_PATH_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_DOMAIN_CONFIG_KEY, TEST_COOKIE_DOMAIN_VALUE)
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_MAX_AGE_CONFIG_KEY, TEST_REFRESH_TOKEN_MAX_AGE_VALUE
    )
    set_key_value_in_config(
        TEST_REFRESH_TOKEN_PATH_CONFIG_KEY, TEST_REFRESH_TOKEN_PATH_KEY_VALUE
    )
    set_key_value_in_config(TEST_COOKIE_SECURE_CONFIG_KEY, "false")

    response_1 = driver_config_app.test_client().get("/login")
    cookies_1 = extract_all_cookies(response_1)
    test_client = driver_config_app.test_client()
    test_client.set_cookie(
        "localhost", "sAccessToken", cookies_1["sAccessToken"]["value"]
    )

    response_2 = test_client.get(
        "/handle", headers={"anti-csrf": response_1.headers.get("anti-csrf")}
    )

    response_dict = json.loads(response_2.data)
    assert "s" in response_dict


def test_custom_response(driver_config_app: Any):
    start_st()

    test_client = driver_config_app.test_client()
    response = test_client.get("/auth/signup/email/exists?email=test@example.com")

    dict_response = json.loads(response.data)
    assert response.status_code == 203
    assert dict_response["custom"]


def test_optional_session(driver_config_app: Any):
    start_st()

    test_client = driver_config_app.test_client()
    response = test_client.get("/handle-session-optional")

    dict_response = json.loads(response.data)
    assert response.status_code == 200
    assert dict_response["s"] == "empty session"


def test_thirdparty_parsing_works(driver_config_app: Any):
    start_st()

    test_client = driver_config_app.test_client()
    data = {
        "state": "afc596274293e1587315c",
        "code": "c7685e261f98e4b3b94e34b3a69ff9cf4.0.rvxt.eE8rO__6hGoqaX1B7ODPmA",
    }
    response = test_client.post("/auth/callback/apple", data=data)

    assert response.status_code == 200
    assert (
        response.data
        == b'<html><head><script>window.location.replace("http://supertokens.io/auth/callback/apple?state=afc596274293e1587315c&code=c7685e261f98e4b3b94e34b3a69ff9cf4.0.rvxt.eE8rO__6hGoqaX1B7ODPmA");</script></head></html>'
    )


from flask.wrappers import Response
from supertokens_python.framework.flask.flask_response import (
    FlaskResponse as SupertokensFlaskWrapper,
)


def test_remove_header_works():
    response = Response()
    st_response = SupertokensFlaskWrapper(response)

    st_response.set_header("foo", "bar")
    assert st_response.get_header("foo") == "bar"
    st_response.remove_header("foo")
    assert st_response.get_header("foo") is None



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Union

from supertokens_python.normalised_url_path import NormalisedURLPath

from .interfaces import (
    CreateEmailVerificationTokenEmailAlreadyVerifiedError,
    CreateEmailVerificationTokenOkResult,
    RecipeInterface,
    RevokeEmailVerificationTokensOkResult,
    UnverifyEmailOkResult,
    VerifyEmailUsingTokenInvalidTokenError,
    VerifyEmailUsingTokenOkResult,
)
from .types import User

if TYPE_CHECKING:
    from supertokens_python.querier import Querier

    from .utils import EmailVerificationConfig


class RecipeImplementation(RecipeInterface):
    def __init__(self, querier: Querier, config: EmailVerificationConfig):
        super().__init__()
        self.querier = querier
        self.config = config

    async def create_email_verification_token(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> Union[
        CreateEmailVerificationTokenOkResult,
        CreateEmailVerificationTokenEmailAlreadyVerifiedError,
    ]:
        data = {"userId": user_id, "email": email}
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/user/email/verify/token"), data
        )
        if "status" in response and response["status"] == "OK":
            return CreateEmailVerificationTokenOkResult(response["token"])
        return CreateEmailVerificationTokenEmailAlreadyVerifiedError()

    async def verify_email_using_token(
        self, token: str, user_context: Dict[str, Any]
    ) -> Union[VerifyEmailUsingTokenOkResult, VerifyEmailUsingTokenInvalidTokenError]:
        data = {"method": "token", "token": token}
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/user/email/verify"), data
        )
        if "status" in response and response["status"] == "OK":
            return VerifyEmailUsingTokenOkResult(
                User(response["userId"], response["email"])
            )
        return VerifyEmailUsingTokenInvalidTokenError()

    async def is_email_verified(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> bool:
        params = {"userId": user_id, "email": email}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/user/email/verify"), params
        )
        return response["isVerified"]

    async def revoke_email_verification_tokens(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> RevokeEmailVerificationTokensOkResult:
        data = {"userId": user_id, "email": email}
        await self.querier.send_post_request(
            NormalisedURLPath("/recipe/user/email/verify/token/remove"), data
        )
        return RevokeEmailVerificationTokensOkResult()

    async def unverify_email(
        self, user_id: str, email: str, user_context: Dict[str, Any]
    ) -> UnverifyEmailOkResult:
        data = {"userId": user_id, "email": email}
        await self.querier.send_post_request(
            NormalisedURLPath("/recipe/user/email/verify/remove"), data
        )
        return UnverifyEmailOkResult()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from typing import TYPE_CHECKING, Union

from supertokens_python.async_to_sync_wrapper import sync
from supertokens_python.framework import BaseResponse

if TYPE_CHECKING:
    from flask import Flask


class Middleware:
    def __init__(self, app: Flask):
        self.app = app
        self.set_before_after_request()
        self.set_error_handler()

    def set_before_after_request(self):
        app = self.app
        from supertokens_python.framework.flask.flask_request import FlaskRequest
        from supertokens_python.framework.flask.flask_response import FlaskResponse
        from supertokens_python.supertokens import manage_session_post_response

        from flask.wrappers import Response

        # There is an error in the typing provided by flask, so we ignore it
        # for now.
        @app.before_request  # type: ignore
        def _():
            from supertokens_python import Supertokens

            from flask import request
            from flask.wrappers import Response

            st = Supertokens.get_instance()

            request_ = FlaskRequest(request)
            response_ = FlaskResponse(Response())

            result: Union[BaseResponse, None] = sync(st.middleware(request_, response_))

            if result is not None:
                if isinstance(result, FlaskResponse):
                    return result.response
                raise Exception("Should never come here")
            return None

        @app.after_request
        def _(response: Response):
            from flask import g

            response_ = FlaskResponse(response)
            if hasattr(g, "supertokens") and g.supertokens is not None:
                manage_session_post_response(g.supertokens, response_)

            return response_.response

    def set_error_handler(self):
        app = self.app
        from supertokens_python.exceptions import SuperTokensError

        from flask import request

        @app.errorhandler(SuperTokensError)
        def _(error: Exception):
            from supertokens_python import Supertokens
            from supertokens_python.framework.flask.flask_request import FlaskRequest
            from supertokens_python.framework.flask.flask_response import FlaskResponse

            from flask.wrappers import Response

            st = Supertokens.get_instance()
            response = Response(json.dumps({}), mimetype="application/json", status=200)

            result: BaseResponse = sync(
                st.handle_supertokens_error(
                    FlaskRequest(request), error, FlaskResponse(response)
                )
            )
            if isinstance(result, FlaskResponse):
                return result.response
            raise Exception("Should never come here")



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Callable, Coroutine, Dict, Union, List, Optional

from supertokens_python.framework.fastapi.fastapi_request import FastApiRequest
from supertokens_python.recipe.session import SessionRecipe
from supertokens_python.types import MaybeAwaitable

from ...interfaces import SessionContainer, SessionClaimValidator


def verify_session(
    anti_csrf_check: Union[bool, None] = None,
    session_required: bool = True,
    override_global_claim_validators: Optional[
        Callable[
            [List[SessionClaimValidator], SessionContainer, Dict[str, Any]],
            MaybeAwaitable[List[SessionClaimValidator]],
        ]
    ] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Callable[..., Coroutine[Any, Any, Union[SessionContainer, None]]]:
    if user_context is None:
        user_context = {}
    from fastapi import Request

    async def func(request: Request) -> Union[SessionContainer, None]:
        baseRequest = FastApiRequest(request)
        recipe = SessionRecipe.get_instance()
        session = await recipe.verify_session(
            baseRequest,
            anti_csrf_check,
            session_required,
            override_global_claim_validators,
            user_context,
        )
        if session is None:
            if session_required:
                raise Exception("Should never come here")
            baseRequest.set_session_as_none()
        else:
            baseRequest.set_session(session)
        return baseRequest.get_session()

    return func



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, Awaitable, Callable, Dict, Union

from supertokens_python.ingredients.emaildelivery import EmailDeliveryInterface
from supertokens_python.recipe.passwordless.emaildelivery.services.backward_compatibility import (
    BackwardCompatibilityService as PlessBackwardCompatibilityService,
)
from supertokens_python.recipe.passwordless.types import (
    CreateAndSendCustomEmailParameters,
)
from supertokens_python.recipe.thirdpartypasswordless.types import EmailTemplateVars
from supertokens_python.supertokens import AppInfo


class BackwardCompatibilityService(EmailDeliveryInterface[EmailTemplateVars]):
    pless_backward_compatiblity_service: PlessBackwardCompatibilityService

    def __init__(
        self,
        app_info: AppInfo,
        create_and_send_custom_email: Union[
            Callable[
                [CreateAndSendCustomEmailParameters, Dict[str, Any]], Awaitable[None]
            ],
            None,
        ] = None,
    ) -> None:
        self.pless_backward_compatiblity_service = PlessBackwardCompatibilityService(
            app_info, create_and_send_custom_email
        )

    async def send_email(
        self, template_vars: EmailTemplateVars, user_context: Dict[str, Any]
    ) -> None:
        await self.pless_backward_compatiblity_service.send_email(
            template_vars, user_context
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from math import ceil
from typing import TYPE_CHECKING, Any, Dict, Union, Optional

from jwt import decode
from supertokens_python.recipe.session.with_jwt.constants import (
    ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY,
)
from supertokens_python.recipe.session.with_jwt.utills import (
    add_jwt_to_access_token_payload,
)
from supertokens_python.utils import get_timestamp_ms

if TYPE_CHECKING:
    from supertokens_python.recipe.openid.interfaces import (
        RecipeInterface as OpenIdRecipeInterface,
    )
    from supertokens_python.recipe.session.interfaces import SessionContainer


def get_session_with_jwt(
    original_session: SessionContainer,
    openid_recipe_implementation: OpenIdRecipeInterface,
) -> SessionContainer:
    original_update_access_token_payload = original_session.update_access_token_payload

    async def update_access_token_payload(
        new_access_token_payload: Optional[Dict[str, Any]],
        user_context: Union[None, Dict[str, Any]] = None,
    ) -> None:
        if user_context is None:
            user_context = {}
        if new_access_token_payload is None:
            new_access_token_payload = {}

        access_token_payload = original_session.get_access_token_payload()

        if ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY not in access_token_payload:
            return await original_update_access_token_payload(
                new_access_token_payload, user_context
            )

        jwt_property_name = access_token_payload[
            ACCESS_TOKEN_PAYLOAD_JWT_PROPERTY_NAME_KEY
        ]

        assert jwt_property_name in access_token_payload
        existing_jwt = access_token_payload[jwt_property_name]

        current_time_in_seconds = ceil(get_timestamp_ms() / 1000)
        decoded_payload: Union[None, Dict[str, Any]] = decode(
            jwt=existing_jwt, options={"verify_signature": False, "verify_exp": False}
        )

        if decoded_payload is None or decoded_payload.get("exp") is None:
            raise Exception("Error reading JWT from session")

        jwt_expiry = 1
        if "exp" in decoded_payload:
            exp = decoded_payload["exp"]
            if exp > current_time_in_seconds:
                # it can come here if someone calls this function well after
                # the access token and the jwt payload have expired. In this case,
                # we still want the jwt payload to update, but the resulting JWT should
                # not be alive for too long (since it's expired already). So we set it to
                # 1 second lifetime.
                jwt_expiry = exp - current_time_in_seconds

        new_access_token_payload = await add_jwt_to_access_token_payload(
            access_token_payload=new_access_token_payload,
            jwt_expiry=jwt_expiry,
            user_id=original_session.get_user_id(),
            jwt_property_name=jwt_property_name,
            openid_recipe_implementation=openid_recipe_implementation,
            user_context=user_context,
        )

        return await original_update_access_token_payload(
            new_access_token_payload, user_context
        )

    original_session.update_access_token_payload = update_access_token_payload
    return original_session
